package ch.usi.inf.sape.tracer.analyzer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeNonSlicedRetrievalSession;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SlicedBytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;

/**
 * An instance of this class will be used as the root
 * object to the trace generated upon running
 * a program on BLAST. We assume, each invocation of
 * BLAST happens in a different process, so there will
 * be only a single instance of this class for each 
 * launching of BLAST.
 * 
 * @author reza
 *
 */
public class RuntimeEnv {


	/**
	 * This set contains a reference to each call stack 
	 * simulator generated for each application thread. 
	 */
	private final Set<ThreadCallStack> threads = new HashSet<>();
	
	/**
	 * Bytecode events generated by all application threads,
	 * ordered by their global clock time during a specific
	 * run. 
	 */
	private final List<BytecodeEvent> bytecodes = new ArrayList<>();

	/**
	 * Provides a unique analysis session, so that the 
	 * results of different analyses don't interfere
	 * with each other.
	 */
	private final AnalysisSessionManager analysisSession = new AnalysisSessionManager();
	
	
	private BytecodeEvent exceptionThrowingInstruction;
	
	/**
	 * This is the app class loader used during instrmentation
	 * and trace phases. We need this for calling back to app
	 * objects during analysis phase.
	 */
	private ClassLoader contextClassLoader;
	
	/**
	 * Application absolute path
	 */
	private String appAbsPath;

	/**
	 * No public constructor to avoid wrong initialization
	 */
	private RuntimeEnv() {}
	
	/**
	 * The singleton instance
	 */
	private static RuntimeEnv instance;
	
	/**
	 * The singleton instance. The first invocation
	 * results in creating a new instance, all the 
	 * other invocations return the already created
	 * one.
	 * 
	 */
	public static RuntimeEnv getInstance(){
		if (instance == null){
			instance = new RuntimeEnv();
		}
		return instance;
	}
	

	/**
	 * @return the bcs
	 */
	public List<BytecodeEvent> getAllBytecodeEvents() {
		return this.bytecodes;
	}
	
	/**
	 * The set of thread call stack simulators corresponding
	 * to each thread used during program execution.
	 * @return
	 */
	public Set<ThreadCallStack> getThreadCallStacks(){
		return this.threads;
	}

	
	public void setExceptionThrowingInstruction(BytecodeEvent op){
		this.exceptionThrowingInstruction = op;
	}
	
	public BytecodeEvent getExceptionThrowingInstruction(){
		return this.exceptionThrowingInstruction;
	}
	
	/**
	 * @return the contextClassLoader
	 */
	public ClassLoader getContextClassLoader() {
		return contextClassLoader;
	}

	/**
	 * @param contextClassLoader the contextClassLoader to set
	 */
	public void setContextClassLoader(ClassLoader contextClassLoader) {
		this.contextClassLoader = contextClassLoader;
	}
	
	/**
	 * Introduces a new thread to this runtime env.
	 * @param tcs
	 */
	public void addThreadCallStack(ThreadCallStack tcs){
		this.threads.add(tcs);
	}
	
	/**
	 * The bytecode events must be always sorted w.r.t 
	 * their global time. In order to satisfy this 
	 * constraint in an efficient way, we perform
	 * a pseudo-insertion sort in this method which
	 * in most of the time would have no shifing 
	 * cost, unless there has been an interleaving
	 * of applicaiton thread between adding an already
	 * generated event to this list and adding a 
	 * new event from another application thread.
	 * This seems quite improbable, so we almost 
	 * never pay the cost for sorting, while having
	 * a sorted list of bytecode event.
	 * @param event
	 * @return
	 */
	public void addNewBytecodeEvent(BytecodeEvent event){
		addNewBytecodeEvent(event, true);
	}

	/**
	 * The bytecode events must be always sorted w.r.t 
	 * their global time. In order to satisfy this 
	 * constraint in an efficient way, we perform
	 * a pseudo-insertion sort in this method which
	 * in most of the time would have no shifing 
	 * cost, unless there has been an interleaving
	 * of applicaiton thread between adding an already
	 * generated event to this list and adding a 
	 * new event from another application thread.
	 * This seems quite improbable, so we almost 
	 * never pay the cost for sorting, while having
	 * a sorted list of bytecode event.
	 * @param event
	 * @param log should the op be logged immediately or postponed
	 */
	public void addNewBytecodeEvent(BytecodeEvent event, boolean log){
		final int size = bytecodes.size();
		int curr = size;
		while (size > 0 && bytecodes.get(curr - 1).getEventTime() > event.getEventTime()){
			System.out.println("Should not reach here!!!!");
			curr--;
		}
		
		//add to global list
		bytecodes.add(curr, event);			 
		
		//add to thread's list
		ThreadCallStack.getCurrentThreadCallStack().addOperation(event, log);		
	}


	/**
	 * @return the analysisSession
	 */
	public AnalysisSessionManager getAnalysisSession() {
		return analysisSession;
	}
	
	//TODO remove this method
	@Deprecated
	public BytecodeEvent getBytecodeEventWithTime(long time){
		 for (BytecodeEvent bytecodeEvent : bytecodes) {
			if (bytecodeEvent.getEventTime() == time)
				return bytecodeEvent;
		}
		 
		throw new IllegalStateException("Event with time "+time+" not found, max time is "+bytecodes.get(bytecodes.size() - 1).getEventTime()); 
	}


	/**
	 * @return the appAbsPath
	 */
	public String getAppAbsPath() {
		return appAbsPath;
	}


	/**
	 * @param appAbsPath the appAbsPath to set
	 */
	public void setAppAbsPath(String appAbsPath) {
		this.appAbsPath = appAbsPath;
	}
	
	
}
