package ch.usi.inf.sape.tracer.test;

import static org.junit.Assert.*;

import java.util.HashSet;
import java.util.Set;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TestWatcher;
import org.junit.runner.Description;

import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.EventConstants;
import ch.usi.inf.sape.tracer.analyzer.QueryFacade;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.RuntimeEnv;
import ch.usi.inf.sape.tracer.analyzer.locations.HeapSpace;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ObjectMetaData;
import ch.usi.inf.sape.tracer.test.snippets.TimeTravelingSnippetCollection;

/**
 * This class contains tests for traveling in the state
 * space of objects and activation records.
 *  
 * An object or activation record during its lifetimes
 * passes through many states by getting different values
 * for its fields/elements/variables. The framework provides
 * the facility to roll back to a particular of the object
 * or the activation record, given a time. 
 * @author reza
 *
 */
public class TimeTravelingTest {

	/**
	 * Logs each test method's name upon execution
	 */
	@Rule
	public TestWatcher testWatcher = new TestWatcher() {
		@Override
		protected void starting(final Description description) {
			String methodName = description.getMethodName();
			System.out.println("Test Method: " + methodName);
		}
	};

	@Test
	public void instanceObjectTest1(){
		Object instance = TimeTravelingSnippetCollection.instanceObjectTest1();
		ObjectMetaData metaData = HeapSpace.getObjectLocationMetaData(instance);
		RuntimeEnv session = RuntimeEnv.getInstance();
		
		
		//The state of the object before checkpoint 2
		BytecodeEvent criterion2 = QueryFacade.findNthLastOccurrenceOfEventClass(3, EventConstants.XSTORE_EVENT, session);
		long time2 = criterion2.getEventTime();
		Binding<? extends MemoryLocation>[] elements2 = metaData.getElementsAtTime(time2);
		Set<String> values2 = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements2) {
			values2.add(binding.getValue().toString());
		}
		
		assertTrue(values2.contains("11.1"));
		assertTrue(values2.contains("111.0"));
		assertTrue(values2.contains("1"));
		assertTrue(values2.contains("NULL"));
		
		//The state of the object before checkpoint 3
		BytecodeEvent criterion3 = QueryFacade.findNthLastOccurrenceOfEventClass(2, EventConstants.XSTORE_EVENT, session);
		long time3 = criterion3.getEventTime();
		Binding<? extends MemoryLocation>[] elements3 = metaData.getElementsAtTime(time3);
		Set<String> values3 = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements3) {
			values3.add(binding.getValue().toString());
		}
		
		assertTrue(values3.contains("11.1"));
		assertTrue(values3.contains("111.0"));
		assertTrue(values3.contains("2"));
		assertFalse(values3.contains("1"));
		assertTrue(values3.contains("[C@1"));
		assertFalse(values3.contains("[C@0"));

		
		//The last state of the object
		BytecodeEvent criterion = QueryFacade.findNthLastOccurrenceOfEventClass(1, EventConstants.XSTORE_EVENT, session);
		long time = criterion.getEventTime();
		Binding<? extends MemoryLocation>[] elements = metaData.getElementsAtTime(time);
		Set<String> values = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements) {
			values.add(binding.getValue().toString());
		}
		
		assertTrue(values.contains("11.1"));
		assertTrue(values.contains("333.0"));
		assertTrue(values.contains("2"));
		assertFalse(values.contains("1"));
		assertTrue(values.contains("[C@1"));
		assertFalse(values.contains("NULL"));

	}
	
	
	@Test
	public void arrayObjectTest1(){
		Object instance = TimeTravelingSnippetCollection.arrayObjectTest1();
		ObjectMetaData metaData = HeapSpace.getObjectLocationMetaData(instance);
		RuntimeEnv session = RuntimeEnv.getInstance();
		
		//The state of the object before checkpoint 3
		BytecodeEvent criterion3 = QueryFacade.findNthLastOccurrenceOfEventClass(1, EventConstants.XSTORE_EVENT, session);
		long time3 = criterion3.getEventTime();
		Binding<? extends MemoryLocation>[] elements3 = metaData.getElementsAtTime(time3);
		Set<String> values3 = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements3) {
			values3.add(binding.getValue().toString());
		}

		assertFalse(values3.contains("1"));
		assertFalse(values3.contains("2"));
		assertFalse(values3.contains("3"));
		assertFalse(values3.contains("4"));
		assertTrue(values3.contains("5"));
		assertTrue(values3.contains("9"));
		
		
		//The state of the object before checkpoint 2
		BytecodeEvent criterion2 = QueryFacade.findNthLastOccurrenceOfEventClass(2, EventConstants.XSTORE_EVENT, session);
		long time2 = criterion2.getEventTime();
		Binding<? extends MemoryLocation>[] elements2 = metaData.getElementsAtTime(time2);
		Set<String> values2 = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements2) {
			values2.add(binding.getValue().toString());
		}
		
		assertTrue(values2.contains("1"));
		assertFalse(values2.contains("2"));
		assertTrue(values2.contains("3"));
		assertFalse(values2.contains("4"));
		assertTrue(values2.contains("5"));

		
		//The state of the object before checkpoint 1
		BytecodeEvent criterion1 = QueryFacade.findNthLastOccurrenceOfEventClass(3, EventConstants.XSTORE_EVENT, session);
		long time1 = criterion1.getEventTime();
		Binding<? extends MemoryLocation>[] elements1 = metaData.getElementsAtTime(time1);
		Set<String> values1 = new HashSet<>();
		for (Binding<? extends MemoryLocation> binding : elements1) {
			values1.add(binding.getValue().toString());
		}
		
		assertTrue(values1.contains("1"));
		assertTrue(values1.contains("2"));
		assertTrue(values1.contains("3"));
		assertTrue(values1.contains("4"));
		assertTrue(values1.contains("5"));

	}
}
