package ch.usi.inf.sape.tracer.test;

import static org.junit.Assert.*;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.WeakHashMap;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TestWatcher;
import org.junit.runner.Description;

import com.sun.source.tree.AssertTree;

import ch.usi.inf.sape.tracer.analyzer.WeakIdentityHashMap2;

/**
 * A weak identity hash map provides a weak mapping
 * from keys to values based on identity checking.
 * 
 * As long as an application object is alive, the 
 * corresponding entry in the map has to remain there.
 * 
 * Equality has to be only based on identity. That 
 * means two string objects with the same value should
 * end up with two different entries in the map.
 * 
 * The value objects have to remain even after the 
 * end of the application.
 * 
 * @author reza
 *
 */
public class WeakIdentityHashMapTest {

	/**
	 * Logs each test method's name upon execution
	 */
	@Rule
	public TestWatcher testWatcher = new TestWatcher() {
		@Override
		protected void starting(final Description description) {
			String methodName = description.getMethodName();
			System.out.println("Test Method: " + methodName);
		}
	};

	@Test
	public void appObjectAliveAfterGCTest(){
		List<Object> objs = new LinkedList<>();
		for (int i = 0; i < 20000; i++) {
			objs.add(new Object());
		}
		
		WeakIdentityHashMap2<Object, Integer> map = new WeakIdentityHashMap2<>();
		for (int i = 0; i < objs.size(); i++) {
			map.put(objs.get(i), i);
		}
		
		//App objects must survive because of the list
		System.gc();
		
		for (int i = 0; i < objs.size(); i++) {
			assertNotNull("The mapping must not be null!", map.get(objs.get(i)));
			assertEquals("The mapping is not done correctly", Integer.valueOf(i), map.get(objs.get(i)));
		}
	}
	
	@Test
	public void appObjectCollectedAfterGCTest(){
		final int SIZE = 200;
		
		WeakIdentityHashMap2<Object, Integer> map = new WeakIdentityHashMap2<>();
		
		for (int i = 0; i < SIZE; i++) {
			map.put(new Object(), i);
		}
		
		//Application objects must be GCed
		System.gc();
		
		/*
		 * A dummy insertion to update the map. According to Map.size() comments,
		 * This result of invoking size() is a snapshot, and may not reflect unprocessed
		 * entries that will be removed before next attempted access
		 * because they are no longer referenced
		 */
		Object o1 = new Object();
		map.put(o1, SIZE + 1);
		for(Object o : map.keySet()){
			System.out.println(o);
		}
		
		assertEquals("All mappings before GC must go away.", 1, map.size());
	}
	
	
	@Test
	public void ObjectIdentityTest1(){
		final int SIZE = 200;
		
		WeakIdentityHashMap2<Object, Integer> map = new WeakIdentityHashMap2<>();
		
		for (int i = 0; i < SIZE; i++) {
			map.put(new String("Hi"), i);
		}
		
		assertEquals("Each instance of String has to have a unique entry.", SIZE, map.size());
	}
	
	@Test
	public void ObjectIdentityTest2(){
		final int SIZE = 200;
		
		WeakIdentityHashMap2<Object, Integer> map = new WeakIdentityHashMap2<>();
		
		for (int i = 0; i < SIZE; i++) {
			map.put("Hi", i);
		}
		
		assertEquals("There has to be only one mapping.", 1, map.size());
	}
}
